/*
 * Copyright (C) 2004-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file string.c memory allocation and string handling functions */

#include "para.h"
#include "string.h"

#include <sys/time.h> /* gettimeofday */
#include <regex.h>
#include <pwd.h>
#include <sys/utsname.h> /* uname() */
#include <string.h>

#include "error.h"

/**
 * paraslash's version of realloc()
 *
 * \param p pointer to the memory block, may be NULL
 * \param size desired new size
 *
 * A wrapper for realloc(3). It calls \p exit(\p EXIT_FAILURE) on errors,
 * i.e. there is no need to check the return value in the caller.
 * \sa realloc(3)
 */
__must_check __malloc void *para_realloc(void *p, size_t size)
{
	/*
	 * No need to check for NULL pointers: If p is NULL, the  call
	 * to realloc is equivalent to malloc(size)
	 */
	if (!(p = realloc(p, size))) {
		PARA_EMERG_LOG("%s", "realloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * paraslash's version of malloc()
 *
 * \param size desired new size
 *
 * A wrapper for malloc(3) which exits on errors.
 * \sa malloc(3)
 */
__must_check __malloc void *para_malloc(size_t size)
{
	void *p = malloc(size);

	if (!p) {
		PARA_EMERG_LOG("%s", "malloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * paraslash's version of calloc()
 *
 * \param size desired new size
 *
 * A wrapper for calloc(3) which exits on errors.
 * \sa calloc(3)
 */
__must_check __malloc void *para_calloc(size_t size)
{
	void *ret = para_malloc(size);

	memset(ret, 0, size);
	return ret;
}

/**
 * paraslash's version of strdup()
 *
 * \param s: string to be duplicated
 *
 * A wrapper for strdup(3). It calls exit(EXIT_FAILURE) on
 * errors, i.e. there is no need to check the return value in the caller.
 * Moreover, this wrapper checks for \a s being NULL and returns an empty
 * string in this case.
 *
 * \sa strdup(3)
 */
__must_check __malloc char *para_strdup(const char *s)
{
	char *ret;

	if ((ret = strdup(s? s: "")))
		return ret;
	PARA_EMERG_LOG("%s", "strdup failed, aborting\n");
	exit(EXIT_FAILURE);
}

/**
 * allocate a sufficiently large string and print into it
 *
 * \param fmt usual format string
 *
 * Produce output according to \a fmt. No artificial bound on the length of the
 * resulting string is imposed. This function either returns a pointer to a
 * string that must be freed by the caller or aborts without returning.
 *
 * \sa printf(3)
 */
__must_check __printf_1_2 __malloc char *make_message(const char *fmt, ...)
{
	char *msg;

	PARA_VSPRINTF(fmt, msg);
	return msg;
}

/**
 * paraslash's version of strcat()
 *
 * \param a string to be appended to
 * \param b string to append
 *
 * Append \a b to \a a. If \a a is NULL, return a copy of \a b, i.e.
 * para_strcat(NULL, b) is equivalent to para_strdup(b).  If \a b is NULL,
 * return \a a without making a copy of \a a.  Otherwise, construct the
 * concatenation \a c, free \a a (but not \a b) and return \a c.
 *
 * \sa strcat(3)
 */
__must_check __malloc char *para_strcat(char *a, const char *b)
{
	char *tmp;

	if (!a)
		return para_strdup(b);
	if (!b)
		return a;
	tmp = make_message("%s%s", a, b);
	free(a);
	return tmp;
}

/**
 * paraslash's version of dirname()
 *
 * \param name pointer to The full path
 *
 * If \a name is \ NULL or the empty string, return \p NULL, Otherwise, Make a
 * copy of \a name and return its directory component. Caller is responsible to
 * free the result.
 */
__must_check __malloc char *para_dirname(const char *name)
{
	char *p, *ret;

	if (!name || !*name)
		return NULL;
	ret = para_strdup(name);
	p = strrchr(ret, '/');
	if (!p)
		*ret = '\0';
	else
		*p = '\0';
	return ret;
}

/**
 * paraslash's version of basename()
 *
 * \param name Pointer to the full path
 *
 * If \a name is \p NULL or the empty string, return \p NULL, Otherwise, make a
 * copy of \a name and return its filename component. Caller is responsible to
 * free the result.
 */
__must_check __malloc char *para_basename(const char *name)
{
	char *p;

	if (!name || !*name)
		return NULL;
	p = strrchr(name, '/');
	if (!p)
		return para_strdup(name);
	p++;
	if (!*p)
		return NULL;
	return para_strdup(p);
}

/**
 * simple search and replace routine
 *
 * \param src source string
 * \param macro_name the name of the macro
 * \param replacement the replacement format string
 *
 * Replace \a macro_name(arg) by \a replacement. \a replacement is a format
 * string which may contain a single string conversion specifier which gets
 * replaced by 'arg'.
 *
 * \return A string in which all matches in \a src are replaced, or NULL if an
 * syntax error was encountered. Caller must free the result.
 *
 * \sa regcomp(3)
 */
__must_check __malloc char *s_a_r(const char *src, const char* macro_name,
		const char *replacement)
{
	regex_t preg;
	size_t  nmatch = 1;
	regmatch_t pmatch[1];
	int eflags = 0;
	char *dest = NULL;
	const char *bufptr = src;

	if (!macro_name || !replacement || !src)
		return para_strdup(src);
	regcomp(&preg, macro_name, 0);
	while (regexec(&preg,  bufptr, nmatch, pmatch, eflags)
			!= REG_NOMATCH) {
		char *tmp, *arg, *o_bracket, *c_bracket;

		o_bracket = strchr(bufptr + pmatch[0].rm_so, '(');
		c_bracket = o_bracket? strchr(o_bracket, ')') : NULL;
		if (!c_bracket)
			goto out;
		tmp = para_strdup(bufptr);
		tmp[pmatch[0].rm_so] = '\0';
		dest = para_strcat(dest, tmp);
		free(tmp);

		arg = para_strdup(o_bracket + 1);
		arg[c_bracket - o_bracket - 1] = '\0';
		tmp = make_message(replacement, arg);
		free(arg);
		dest = para_strcat(dest, tmp);
		free(tmp);
		bufptr = c_bracket;
		bufptr++;
	}
	dest = para_strcat(dest, bufptr);
//	PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
out:
	regfree(&preg);
	return dest;
}

/**
 * replace a string according to a list of macros
 *
 * \param macro_list the array containing a macro/replacement pairs.
 * \param src the source string
 *
 * This function just calls s_a_r() for each element of \a macro_list.
 */
__must_check __malloc char *s_a_r_list(struct para_macro *macro_list, char *src)
{
	struct para_macro *mp = macro_list;
	char *ret = NULL, *tmp = para_strdup(src);

	while (mp->name) {
		ret = s_a_r(tmp, mp->name, mp->replacement);
		free(tmp);
		if (!ret) /* syntax error */
			return NULL;
		tmp = ret;
		mp++;
	}
	//PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
	return ret;
}

/**
 * cut trailing newline
 *
 * \param buf the string to be chopped.
 *
 * Replace the last character in \a buf by zero if it is euqal to
 * the newline character.
 */
void chop(char *buf)
{
	int n = strlen(buf);
	if (!n)
		return;
	if (buf[n - 1] == '\n')
		buf[n - 1] = '\0';
}

/**
 * get a random filename
 *
 * This is by no means a secure way to create temporary files in a hostile
 * direcory like /tmp. However, it is OK to use for temp files, fifos, sockets
 * that are created in ~/.paraslash. Result must be freed by the caller.
 */
__must_check __malloc char *para_tmpname(void)
{
	struct timeval now;
	gettimeofday(&now, NULL);
	srand(now.tv_usec);
	return make_message("%08i", rand());
}

/**
 * create unique temporary file
 *
 * \param template the template to be passed to mkstemp()
 * \param mode the desired mode of the tempfile
 *
 * This wrapper for mkstemp additionally uses fchmod() to
 * set the given mode of the tempfile if mkstemp() returned success.
 * Return value: The file descriptor of the temp file just created on success.
 * On errors, -E_MKSTEMP or -E_FCHMOD is returned.
 */
__must_check int para_mkstemp(char *template, mode_t mode)
{
	int tmp, fd = mkstemp(template);

	if (fd < 0)
		return -E_MKSTEMP;
	tmp = fchmod(fd, mode);
	if (tmp >= 0)
		return fd;
	close(fd);
	unlink(template);
	return -E_FCHMOD;
}

/**
 * get the logname of the current user
 *
 * \return A dynammically allocated string that must be freed by the caller. On
 * errors, the string "unknown user" is returned, i.e. this function never
 * returns NULL.
 */
__must_check __malloc char *para_logname(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_name : "unknown_user");
}

/**
 * get the home directory of the current user
 *
 * \return A dynammically allocated string that must be freed by the caller. If
 * the home directory could not be found, this function returns "/tmp".
 */
__must_check __malloc char *para_homedir(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_dir : "/tmp");
}

/**
 * split string and return pointers to its parts.
 *
 * \param args the string to be split
 * \param argv_ptr  pointer to the list of substrings
 * \param delim delimiter
 *
 * This function modifies \a args by replacing each occurance of \a delim by
 * zero. A NULL-terminated array of pointers to char* is allocated dynamically
 * and these pointers are initialized to point to the broken-up substrings
 * within \a args. A pointer to this array is returned via \a argv_ptr. It's OK
 * to call this function with \a args == NULL.
 *
 * \return The number of substrings found in \a args.
 */

__must_check unsigned split_args(char *args, char ***argv_ptr, const char *delim)
{
	char *p = args;
	char **argv;
	size_t n = 0, i, j;

	while ((i = strcspn(p, delim)) && (p += i)) {
		p += strspn(p, delim);
		n++;
	}
	*argv_ptr = para_malloc((n + 1) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args;
	while (p) {
		argv[i] = p;
		j = strcspn(p, delim);
		if (!j)
			break;
		p += strcspn(p, delim);
		if (*p) {
			*p = '\0';
			p++;
			p += strspn(p, delim);
		}
		i++;
	}
	argv[n] = NULL;
	return n;
}

/**
 * ensure that file descriptors 0, 1, and 2 are valid
 *
 * Common approach that opens /dev/null until it gets a file descriptor greater
 * than two.
 *
 * \sa okir's Black Hats Manual.
 */
void valid_fd_012(void)
{
	while (1) {
	int     fd;

		fd = open("/dev/null", O_RDWR);
		if (fd < 0)
			exit(EXIT_FAILURE);
		if (fd > 2) {
			close(fd);
			break;
		}
	}
}

/**
 * get the own hostname
 *
 * \return A dynammically allocated string containing the hostname.
 *
 * \sa uname(2)
 */
__malloc char *para_hostname(void)
{
	struct utsname u;

	uname(&u);
	return para_strdup(u.nodename);
}
