/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include "para.h"
#include "string.h"
#include "write.h"
#include "error.h"

const char *writer_names[] ={WRITER_NAMES};
struct writer writers[NUM_SUPPORTED_WRITERS] = {WRITER_ARRAY};

int wng_write(struct writer_node_group *g, char *buf, size_t *loaded)
{
	int ret, i, need_more_writes = 1;
	size_t min_written = 0;

	while (need_more_writes) {
		need_more_writes = 0;
		FOR_EACH_WRITER_NODE(i, g) {
			size_t w = g->written[i];
			int bytes_to_write;
			struct writer_node *wn = &g->writer_nodes[i];
			if (!i)
				min_written = w;
			else
				min_written = PARA_MIN(min_written, w);
			if (w == *loaded)
				continue;
			if (!g->eof && (*loaded < wn->chunk_bytes + w))
				continue;
			bytes_to_write = PARA_MIN(wn->chunk_bytes,
				*loaded - w);
			ret = wn->writer->write(buf + w, bytes_to_write, wn);
			if (ret < 0)
				goto out;
			if (ret != bytes_to_write)
				PARA_WARNING_LOG("short write: %d/%d\n", ret,
					bytes_to_write);
			g->written[i] += ret;
			need_more_writes = 1;
		}
	}
	*loaded -= min_written;
	ret = 0;
	if (g->eof)
		goto out;
	if (*loaded)
		memmove(buf, buf + min_written, *loaded);
	FOR_EACH_WRITER_NODE(i, g)
		g->written[i] -= min_written;
	ret = 1;
out:
	return ret;
}

int wng_open(struct writer_node_group *g)
{
	int i, ret = 1;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		ret = wn->writer->open(wn);
		if (ret < 0)
			goto out;
		wn->chunk_bytes = ret;
		g->max_chunk_bytes = PARA_MAX(g->max_chunk_bytes, ret);
	}
out:
	return ret;
}

void wng_close(struct writer_node_group *g)
{
	int i;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		wn->writer->close(wn);
	}
}

struct writer_node_group *wng_new(unsigned num_writers)
{
	struct writer_node_group *g = para_calloc(sizeof(struct writer_node_group));
	g->num_writers = num_writers;
	g->writer_nodes = para_calloc(num_writers
		* sizeof(struct writer_node));
	g->written = para_calloc(num_writers * sizeof(size_t));
	return g;
}

void wng_destroy(struct writer_node_group *g)
{
	if (!g)
		return;
	free(g->written);
	free(g->writer_nodes);
	free(g);
}

void init_supported_writers(void)
{
	int i;

	FOR_EACH_WRITER(i)
		writers[i].init(&writers[i]);
}

int check_writer_arg(const char *arg)
{
	int i, ret = -E_WRITE_COMMON_SYNTAX;
	char *a = para_strdup(arg), *p = strchr(a, ':');
	if (p)
		*p = '\0';
	p++;
	FOR_EACH_WRITER(i) {
		if (strcmp(writer_names[i], a))
			continue;
		ret = i;
		goto out;
	}
out:
	free(a);
	return ret;
}

struct writer_node_group *setup_default_wng(void)
{
	struct writer_node_group *wng = wng_new(1);
	enum writer_enum default_writer;

	if (NUM_SUPPORTED_WRITERS == 1)
		default_writer = FILE_WRITE;
	else
		default_writer = 1;
	wng->writer_nodes[0].writer = &writers[default_writer];
	PARA_INFO_LOG("using default writer: %s\n",
		writer_names[default_writer]);
	return wng;
}
